@@ -1,81 +1,115 @@
-# a wrapper for memuc.exe as a library to control virual machines
-# documentation detailed in memuc_documentation.md
+"""a wrapper for memuc.exe as a library to control virual machines"""
 
+import contextlib
+import re
+from os import environ
 from os.path import abspath, expanduser, expandvars, join, normpath
-from subprocess import PIPE, Popen
-from typing import Literal
+from subprocess import PIPE, CalledProcessError, Popen, TimeoutExpired, check_output
+from typing import Literal, Union
 
-from pymemuc.types import VMInfo
+from pymemuc.exceptions import PyMemucError, PyMemucIndexError, PyMemucTimeoutExpired
+from pymemuc.vminfo import VMInfo
 
+# check for debug mode
+DEBUG = environ.get("PYTHON_ENV") == "development"
+
+# check for windows registry support
 try:
     from winreg import HKEY_LOCAL_MACHINE, ConnectRegistry, OpenKey, QueryValueEx
 
-    windows_registry = True
+    WINREG_EN = True
 except ImportError:
-    print(
-        "Windows Registry is not supported on this platform, you must specify the path to memuc.exe manually"
-    )
-    windows_registry = False
+    if DEBUG:
+        print(
+            "Windows Registry is not supported on this platform, you must specify the path to memuc.exe manually"
+        )
+    WINREG_EN = False
+
+if __name__ == "__main__" and DEBUG:
+    print("Debug mode enabled")  # debug
 
 
 class PyMemuc:
     """A class to interact with the memuc.exe command line tool to control virtual machines.
 
-    :param memuc_path: Path to memuc.exe. If windows registry is supported, this is not necessary. Defaults to None.
+    :param memuc_path: Path to memuc.exe. Set to None for autodetect.
     :type memuc_path: str, optional
     """
 
-    def __init__(self, memuc_path=None) -> None:
-        """initialize the class, automatically finding memuc.exe if windows registry is supported, otherwise a path must be specified"""
-        if windows_registry:
-            self.memuc_path = join(self._get_memu_top_level(), "memuc.exe")
+    def __init__(self, memuc_path: Union[str, None] = None) -> None:
+        """initialize the class, automatically finding memuc.exe if windows registry is supported,
+        otherwise a path must be specified"""
+        if WINREG_EN:
+            self.memuc_path: str = join(self._get_memu_top_level(), "memuc.exe")
         elif memuc_path is not None:
+            self.memuc_path: str = memuc_path
+        else:
             raise PyMemucError(
                 "Windows Registry is not supported on this platform, you must specify the path to memuc.exe manually"
             )
-        else:
-            self.memuc_path = memuc_path
 
     def _get_memu_top_level(self) -> str:
         """locate the path of the memu directory using windows registry keys
 
         :return: the path of the memu directory
         :rtype: str
+        :raises PyMemucError: an error if memu is not installed
         """
-        try:
-            akey = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\MEmu"
-            areg = ConnectRegistry(None, HKEY_LOCAL_MACHINE)
-            akey = OpenKey(areg, akey)
-        except FileNotFoundError:
-            try:
-                akey = r"SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall\MEmu"
-                areg = ConnectRegistry(None, HKEY_LOCAL_MACHINE)
-                akey = OpenKey(areg, akey)
-            except FileNotFoundError as e:
-                raise PyMemucError("MEmuc not found, is it installed?") from e
-        return str(join(normpath(QueryValueEx(akey, "InstallLocation")[0]), "Memu"))
 
-    def run(self, args, non_blocking=False) -> tuple[int, str]:
+        for key in [  # keys to search for memu
+            r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\MEmu",
+            r"SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall\MEmu",
+        ]:
+            with contextlib.suppress(FileNotFoundError):
+                akey = OpenKey(ConnectRegistry(None, HKEY_LOCAL_MACHINE), key)
+                return str(
+                    join(normpath(QueryValueEx(akey, "InstallLocation")[0]), "Memu")
+                )
+        raise PyMemucError("MEmuc not found, is it installed?")
+
+    def run(self, args: list[str], non_blocking=False) -> tuple[int, str]:
         """run a command with memuc.exe
 
         :param args: a list of arguments to pass to memuc.exe
-        :type args: list
+        :type args: list[str]
         :param non_blocking: whether to run the command in the background. Defaults to False.
         :type non_blocking: bool, optional
         :return: the return code and the output of the command
         :rtype: tuple[int, str]
+        :raises PyMemucError: an error if the command failed
         """
         args.insert(0, self.memuc_path)
         args += "-t" if non_blocking else ""
-        p = Popen(args, stdout=PIPE, shell=False)
+        with Popen(args, stdout=PIPE, shell=False) as proc:
+            (out, err) = proc.communicate()
+            out = out.decode("utf-8")  # convert bytes to string
+            status = proc.wait()
+            if err:
+                raise PyMemucError(err)
+            if DEBUG:
+                # print the command that was run and the output for debugging
+                print(f"Command: {' '.join(args)}\nOutput [{status}]: {out}")  # debug
+            return status, out
 
-        (output, err) = p.communicate()
-        output = output.decode("utf-8")  # convert bytes to string
-        p_status = p.wait()
-        if err:
-            raise PyMemucError(err)
-        # print(f"Command output [{p_status}]: {output}") # debug
-        return p_status, output
+    def run_with_timeout(self, args: list[str], timeout=10) -> tuple[int, str]:
+        """run a command with memuc.exe with a timeout
+
+        :param args: a list of arguments to pass to memuc.exe
+        :type args: list[str]
+        :param timeout: the timeout in seconds. Defaults to 10.
+        :type timeout: int, optional
+        :return: the return code and the output of the command
+        :rtype: tuple[int, str]
+        :raises PyMemucError: an error if the command failed
+        :raises PyMemucTimeoutExpired: an error if the command timed out
+        """
+        args.insert(0, self.memuc_path)
+        try:
+            return (1, check_output(args, timeout=timeout).decode("utf-8"))
+        except CalledProcessError as err:
+            raise PyMemucError(err) from err
+        except TimeoutExpired as err:
+            raise PyMemucTimeoutExpired(err) from err
 
     def create_vm(self, vm_version="76") -> int:
         """Create a new VM
@@ -90,13 +124,11 @@ class PyMemuc:
         success = status == 0 and output is not None and "SUCCESS" in output
         if not success:
             raise PyMemucError(f"Failed to create VM: {output}")
-        # filter output with regex r"index:(\w)"
-        import re
-
         try:
-            return int(re.search(r"index:(\w)", output).group(1)) # type: ignore
-        except AttributeError: return -1
-
+            indecies = re.search(r"index:(\w)", output)
+            return -1 if indecies is None else int(indecies[1])
+        except AttributeError:
+            return -1
 
     def delete_vm(self, vm_index=None, vm_name=None) -> Literal[True]:
         """Delete a VM, must specify either a vm index or a vm name
@@ -165,10 +197,9 @@ class PyMemuc:
             return self.run(
                 ["-i", str(vm_index), "export", f'"{file_name}"'], non_blocking
             )
-        elif vm_name is not None:
+        if vm_name is not None:
             return self.run(["-n", vm_name, "export", f'"{file_name}"'], non_blocking)
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
 
     def import_vm(self, file_name="vm.ova", non_blocking=False) -> Literal[True]:
         """Import a VM from a file
@@ -277,38 +308,39 @@ class PyMemuc:
         """
 
         if vm_index is not None:
-            status, output = self.run(
+            _, output = self.run(
                 [
                     "-i",
                     str(vm_index),
                     "listvms",
-                    "-running" if running else "",
+                    "-r" if running else "",
                     "-s" if disk_info else "",
                 ]
             )
         elif vm_name is not None:
-            status, output = self.run(
+            _, output = self.run(
                 [
                     "-n",
                     vm_name,
                     "listvms",
-                    "-running" if running else "",
+                    "-r" if running else "",
                     "-s" if disk_info else "",
                 ]
             )
         else:
-            status, output = self.run(
-                ["listvms", "-running" if running else "", "-s" if disk_info else ""]
+            _, output = self.run(
+                ["listvms", "-r" if running else "", "-s" if disk_info else ""]
             )
 
         output = output.split("\r\n")
         parsed_output = []
 
         # parse the output into a list of dictionaries representing the VMs
-        # output will contain a list of vm values seperated by commas, if disk_info is True, each vm will have 6 values, otherwise 5
-        for vm in output:
-            if vm:
-                vm_info = vm.split(",")
+        # output will contain a list of vm values seperated by commas
+        # if disk_info is True, each vm will have 6 values, otherwise 5
+        for vm_str in output:
+            if vm_str:
+                vm_info = vm_str.split(",")
                 parsed_output.append(
                     {
                         "index": int(vm_info[0]),
@@ -329,7 +361,7 @@ class PyMemuc:
         :return: True if the VM is running, False otherwise
         :rtype: bool
         """
-        status, output = self.run(["-i", str(vm_index), "isrunning"])
+        _, output = self.run(["-i", str(vm_index), "isrunning"])
         return "Running" in output
 
     def sort_out_all_vm(self) -> bool:
@@ -403,7 +435,7 @@ class PyMemuc:
         :return: the return code and the output of the command.
         :rtype: tuple[int, str]
         """
-        return self.run("taskstatus", task_id)
+        return self.run(["taskstatus"], task_id)
 
     def get_configuration_vm(self, config_key, vm_index=None, vm_name=None) -> str:
         """Get a VM configuration, must specify either a vm index or a vm name
@@ -655,7 +687,7 @@ class PyMemuc:
             raise PyMemucError(f"Failed to connect internet: {output}")
         return True
 
-    def disconnect_internet_vm(self, vm_index=None, vm_name=None):
+    def disconnect_internet_vm(self, vm_index=None, vm_name=None) -> Literal[True]:
         """Disconnect the internet on a VM, must specify either a vm index or a vm name
 
         :param vm_index: VM index. Defaults to None.
@@ -738,10 +770,9 @@ class PyMemuc:
         """
         if vm_index is not None:
             return self.run(["-i", str(vm_index), "execcmd", f'"{command}"'])
-        elif vm_name is not None:
+        if vm_name is not None:
             return self.run(["-n", vm_name, "execcmd", f'"{command}"'])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
 
     def change_gps_vm(
         self, latitude: float, longitude: float, vm_index=None, vm_name=None
@@ -776,7 +807,7 @@ class PyMemuc:
         return True
 
     # TODO: fix parsing of the output
-    def get_public_ip_vm(self, vm_index=None, vm_name=None):
+    def get_public_ip_vm(self, vm_index=None, vm_name=None) -> tuple[int, str]:
         """Get the public IP of a VM, must specify either a vm index or a vm name
 
         :param vm_index: VM index. Defaults to None.
@@ -791,10 +822,9 @@ class PyMemuc:
             return self.run(
                 ["-i", str(vm_index), 'execcmd "wget -O- whatismyip.akamai.com"']
             )
-        elif vm_name is not None:
+        if vm_name is not None:
             return self.run(["-n", vm_name, 'execcmd "wget -O- whatismyip.akamai.com"'])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
 
     def zoom_in_vm(self, vm_index=None, vm_name=None) -> Literal[True]:
         """Zoom in on a VM, must specify either a vm index or a vm name
@@ -840,44 +870,51 @@ class PyMemuc:
             raise PyMemucError(f"Failed to zoom in: {output}")
         return True
 
-    def get_app_info_list_vm(self, vm_index=None, vm_name=None) -> list[str]:
+    def get_app_info_list_vm(
+        self, vm_index=None, vm_name=None, timeout=10
+    ) -> list[str]:
         """Get the list of apps installed on a VM, must specify either a vm index or a vm name
 
         :param vm_index: VM index. Defaults to None.
         :type vm_index: int, optional
         :param vm_name: VM name. Defaults to None.
         :type vm_name: str, optional
+        :param timeout: Timeout for the command. Defaults to 10.
+        :type timeout: int, optional
         :raises PyMemucIndexError: an error if neither a vm index or a vm name is specified
         :return: the list of packages installed on the VM
         :rtype: list[str]
         """
-        if vm_index is not None:
-            status, output = self.run(["-i", str(vm_index), "getappinfolist"])
-        elif vm_name is not None:
-            status, output = self.run(["-n", vm_name, "getappinfolist"])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
-        output = output.split("\r\n")
-        output = [line.replace("package:", "") for line in output if line != ""]
-        return output
+        try:
+            if vm_index is not None:
+                _, output = self.run_with_timeout(
+                    ["-i", str(vm_index), "getappinfolist"],
+                    timeout=timeout,
+                )
+            elif vm_name is not None:
+                _, output = self.run_with_timeout(
+                    ["-n", vm_name, "getappinfolist"],
+                    timeout=timeout,
+                )
+            else:
+                raise PyMemucIndexError("Please specify either a vm index or a vm name")
+            output = output.split("\r\n")
+            output = [line.replace("package:", "") for line in output if line != ""]
+            return output
+        except PyMemucTimeoutExpired:
+            return []
 
     # TODO: debug this, it doesn't work
     def set_accelerometer_vm(
         self,
-        x: float,
-        y: float,
-        z: float,
+        value: tuple[float, float, float],
         vm_index=None,
         vm_name=None,
-    ):
+    ) -> tuple[int, str]:
         """Set the accelerometer on a VM, must specify either a vm index or a vm name
 
-        :param x: X value
-        :type x: float
-        :param y: Y value
-        :type y: float
-        :param z: Z value
-        :type z: float
+        :param value: the accelerometer value to set
+        :type value: tuple[float, float, float]
         :param vm_index: VM index. Defaults to None.
         :type vm_index: int, optional
         :param vm_name: VM name. Defaults to None.
@@ -886,22 +923,37 @@ class PyMemuc:
         :return: the return code and the output of the command.
         :rtype: tuple[int, str]
         """
+
         if vm_index is not None:
             return self.run(
-                ["-i", str(vm_index), "accelerometer", str(x), str(y), str(z)]
+                [
+                    "-i",
+                    str(vm_index),
+                    "accelerometer",
+                    str(value[0]),
+                    str(value[1]),
+                    str(value[2]),
+                ]
             )
-        elif vm_name is not None:
-            return self.run(["-n", vm_name, "accelerometer", str(), str(y), str(z)])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
+        if vm_name is not None:
+            return self.run(
+                [
+                    "-n",
+                    vm_name,
+                    "accelerometer",
+                    str(value[0]),
+                    str(value[1]),
+                    str(value[2]),
+                ]
+            )
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
 
-    # TODO: functionally works, but hangs after the command is run
     def create_app_shortcut_vm(
         self,
         package_name: str,
         vm_index=None,
         vm_name=None,
-    ):
+    ) -> tuple[int, str]:
         """Create an app shortcut on a VM, must specify either a vm index or a vm name
 
         :param package_name: Package name
@@ -911,18 +963,24 @@ class PyMemuc:
         :param vm_name: VM name. Defaults to None.
         :type vm_name: str, optional
         :raises PyMemucIndexError: an error if neither a vm index or a vm name is specified
+        :raises PyMemucTimeoutExpired: an error if the command times out
         :return: the return code and the output of the command.
         :rtype: tuple[int, str]
         """
         if vm_index is not None:
-            return self.run(["-i", str(vm_index), "createshortcut", package_name])
-        elif vm_name is not None:
-            return self.run(["-n", vm_name, "createshortcut", package_name])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
+            return self.run_with_timeout(
+                ["-i", str(vm_index), "createshortcut", package_name]
+            )  # can raise timeout
+        if vm_name is not None:
+            return self.run_with_timeout(
+                ["-n", vm_name, "createshortcut", package_name]
+            )  # can raise timeout
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
 
     # TODO: parse the output
-    def send_adb_command_vm(self, command, vm_index=None, vm_name=None):
+    def send_adb_command_vm(
+        self, command, vm_index=None, vm_name=None
+    ) -> tuple[int, str]:
         """Send an ADB command to a VM, must specify either a vm index or a vm name
 
         :param command: ADB command
@@ -937,30 +995,6 @@ class PyMemuc:
         """
         if vm_index is not None:
             return self.run(["-i", str(vm_index), "adb", f'"{command}"'])
-        elif vm_name is not None:
+        if vm_name is not None:
             return self.run(["-n", vm_name, "adb", f'"{command}"'])
-        else:
-            raise PyMemucIndexError("Please specify either a vm index or a vm name")
-
-
-class PyMemucError(Exception):
-    """PyMemuc error class"""
-
-    def __init__(self, value):
-        self.value = value
-
-    def __str__(self):
-        return repr(self.value)
-
-
-class PyMemucIndexError(PyMemucError):
-    """PyMemuc index error class"""
-
-    def __init__(self, value):
-        self.value = value
-
-    def __str__(self):
-        return repr(self.value)
-
-
-
+        raise PyMemucIndexError("Please specify either a vm index or a vm name")
